#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "image.h"

#define STATE_HEADER 1
#define STATE_WIDTH  2
#define STATE_HEIGHT 3
#define STATE_DATA   4

int conv_state=STATE_HEADER;
int conv_number=0;
int conv_data_byte_pos=0;


#define _IS_NEWLINE( c ) ( c == '\n' )
#define _IS_SPACE( c ) ( c == ' ' )

void pbm_image_initx( struct pbm_image *img )
{
	memset( img, 0, sizeof( struct pbm_image ) );
	conv_state = STATE_HEADER;
}


int pgm_image_read_byte( struct pbm_image *img, const char byte )
{
	switch ( conv_state )
	{
	case STATE_HEADER:
		if ( _IS_NEWLINE( byte ) )
		{
			conv_state = STATE_WIDTH;
		}
		break;
	case STATE_WIDTH:
		if ( _IS_SPACE( byte ) )
		{
			(img->width) = conv_number;
			conv_number=0;
			conv_state=STATE_HEIGHT;
		}
		else
		{
			conv_number*=10;
			conv_number+=(byte - '0');
		}
		break;
	case STATE_HEIGHT:
		if ( _IS_NEWLINE( byte ) )
		{
			(img->height) = conv_number;
			conv_number=0;
			conv_state=STATE_DATA;
			(img->data) = (char *)malloc( sizeof(char) * (img->width) * (img->height) );

			if ( !(img->data) )
			{
				return ( -1 );
			}
		}
		else
		{
			conv_number*=10;
			conv_number+=(byte - '0');
		}
		break;
	case STATE_DATA:
		if ( _IS_SPACE( byte ) )
		{
			conv_data_byte_pos++;
		}
		else if ( _IS_NEWLINE( byte ) )
		{
			/* do nothing */
		}
		else
		{
			(img->data)[conv_data_byte_pos] = (byte - '0');
		}
		break;
	}

	return ( 0 );
}

int pbm_image_from_file( struct pbm_image *img, const char *filename )
{
	int fd;
#	define BUF_SIZE 1024
	char buf[BUF_SIZE+1];
	size_t readb;
	char *ptr;
	char *end;

	/* open file */
	fd = open( filename, O_RDONLY, 0 );

	if ( fd < 0 )
	{
		printf( "can't open file: %s\n", filename );
		return ( -1 );
	}

	while ( (readb = read( fd, &buf, BUF_SIZE )) > 0 )
	{
		ptr = &buf[0];
		end = &buf[readb];

		while ( ptr < end )
		{
			if ( pgm_image_read_byte( img, *ptr ) < 0 )
			{
				printf( "error reading pbm file: oom\n" );
				close( fd );
				return ( -1 );
			}
			ptr++;
		}
	}

	close( fd );

	return ( 0 );
}
#include <time.h>
int pbm_image_get_rand_black_point( struct pbm_image *img, int *width, int *height )
{
	int max_nmbr;
	int current;
	int random;

	max_nmbr=(img->width) * (img->height);
//	srand( time( NULL ) );
	do
	{
        random=rand()*rand();
		current = (random % max_nmbr);
	} while ( (img->data)[current] != 1 );

	if ( height )
	{
		(*height) = current / (img->width);
	}
	if ( width )
	{
		(*width) = current % (img->width);
	}

	return ( current );
}

int pbm_image_is_black_point( struct pbm_image *img, int width, int height )
{
    int data_pixel;
    data_pixel = height * (img->width) + width;
    
    return ( img->data[data_pixel] == 1 );
}
